/* 
 3D camera rotator - no translation
 -Press 'n' to toggle looking around on/off
 
 IMPORTANT: Since this script uses the java.awt.robot class, it will require a signed certificate in
 order to run within a web browser. To avoid this, use z_3Drotate1b.
 
 Zachary Seldess, 6/25/08 [http://www.zacharyseldess.com]
*/

import processing.opengl.*;
import java.awt.*;
import javax.swing.SwingUtilities;

//---CHANGE THESE VARIABLES TO YOUR LIKING---//
float cam[] = {0., 0., 0.}; // set camera coordinates
float lookat[] = {0., 0., 1.}; // set lookat coordinates
float lookYangle = PI*0.9; // lookat angle constraints (PI == 180 degree rotation, PI*0.5 == 90 degrees)
// NOTE: setting lookYangle to PI or greater won't look good. I recommend not exceeding PI*0.9
float easing = 0.05; // set smoothing (0.01 == very smooth, 1. == no smoothing)
int mouseTerritory[] = {800, 600}; // set the pixel boundaries of the cursor - effects rotation speed
//-------------------------------------------//

int mouseMovement = 0;
boolean centering = false;
boolean mouseListen = false;
boolean hideCursor = false;
int n = 0;
float easeX;
float easeY;
int lastX;
int lastY;
float mX;
float mY;
int sizeX;
int sizeY;
int sizeXedge;
int sizeYedge;

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////

//---functions called during setup()---//
void navSetup() {
  sizeX = width;
  sizeY = height;
  checkScreenSize();
  sizeXedge = sizeX/10;
  sizeYedge = sizeY/10;
  deriveCursor();
  setCursor(lastX, lastY);
  checkRotation();
}
//--------------------------------------//


//---functions called only once during navSetup()---//
void checkScreenSize() {
  if (sizeX >= mouseTerritory[0]) {
    sizeX = mouseTerritory[0];
  }
  if (sizeY >= mouseTerritory[1]) {
    sizeY = mouseTerritory[1];
  }
}

void deriveCursor() {
  float x = lookat[0]-cam[0];
  float y = lookat[1]-cam[1];
  float z = lookat[2]-cam[2];  
  float hyp = mag(x,y,z);
  float sine = y/hyp;
  float arcsine = asin(sine);
  float Ymouse = map(arcsine, -lookYangle/2, lookYangle/2, sizeYedge, sizeY-sizeYedge);
  float atangent = atan2(z,x);
  float Xmouse = map(atangent, 0, TWO_PI, sizeXedge, sizeX-sizeXedge);
  if (Xmouse < sizeXedge) {
    Xmouse = (sizeX-sizeXedge) - (sizeXedge-Xmouse);
  }
  if (Xmouse > sizeX-sizeXedge) {
    Xmouse = (Xmouse-(sizeX-sizeXedge)) + sizeXedge;
  }
  if (Ymouse < sizeYedge) {
    Ymouse = float(sizeYedge);
    println("lookat vector exceeds bounds of specified lookYangle: " + lookYangle);
    println("constraining lookat accordingly...");
  }
  if (Ymouse > sizeY-sizeYedge) {
    Ymouse = float(sizeY-sizeYedge);
    println("lookat vector exceeds bounds of specified lookYangle: " + lookYangle);
    println("constraining lookat accordingly...");
  }
  mouseX = int(Xmouse);
  mouseY = int(Ymouse);
  pmouseX = int(Xmouse);
  pmouseY = int(Ymouse);
  lastX = int(Xmouse);
  lastY = int(Ymouse);
  easeX = int(Xmouse);
  easeY = int(Ymouse);
}
//--------------------------------------//


//--- functions called during draw() ---//
void navUpdate() {
  keyActivity();
  mouseState();
  if (n == 2) {
    mouseEase();
    checkRotation();
  }
  camera(cam[0], cam[1], cam[2], lookat[0], lookat[1], lookat[2], 0, 1, 0);
}

void keyActivity() {
  if (keyPressed) {
    if (key == 'n') {
      if (n == 0) {
        n = 1;
        noCursor();
        mouseX = lastX;
        mouseY = lastY;
        pmouseX = lastX;
        pmouseY = lastY;
        setCursor(lastX, lastY);
        mouseListen = true;
      }
      if (n == 2) {
        n = 3;
        lastX = mouseX;
        lastY = mouseY;
        cursor();
        mouseListen = false;
        hideCursor = false;
      }
    }
  }
  else if (!keyPressed) {
    if (key == 'n') {
      if (n == 1) {
        n = 2;
      }
      else if (n == 3) {
        n = 0;
      }
    }
  }
}

void mouseState() {
  if (mouseListen) {
    if (centering) { // don't add the robot mouse movement to 'mouseMovement'
      centering = false;
    }
    else { // add normal mouse movement to 'mouseMovement'
      checkCursor();
    }
  }
}

void mouseEase() {  
  float dx = mouseX - easeX;
  if (mouseX < sizeXedge*2 && pmouseX > sizeX-(sizeXedge*2)) {
    easeX = easeX - (sizeX-(sizeXedge*2));
    dx = mouseX - easeX;
  }
  if (mouseX > sizeX-(sizeXedge*2) && pmouseX < sizeXedge*2) {
    easeX = easeX + (sizeX-(sizeXedge*2));
    dx = mouseX - easeX;
  }
  if(abs(dx) > 1) {
    easeX += dx * easing;
  }
  float dy = mouseY - easeY;
  if(abs(dy) > 1) {
    easeY += dy * easing;
  }
}

void checkRotation() {
  mX = ((easeX-sizeXedge)/float(sizeX-(sizeXedge*2)))*TWO_PI; // scale mouseX movement between 0 and 2*PI
  mY = ((easeY-sizeYedge)/float(sizeY-(sizeYedge*2)))*lookYangle - lookYangle/2;
  lookat[0] = cam[0] + cos(mX)*cos(mY);
  lookat[1] = cam[1] + sin(mY);
  lookat[2] = cam[2] + sin(mX)*cos(mY);
}

void checkCursor() {
  mouseMovement = mouseX - pmouseX;
  if (mouseMovement < 0) {
    if (mouseX < sizeXedge) {
      moveCursorR(mouseY);
    }
  }
  if (mouseMovement > 0) {
    if (mouseX > sizeX-sizeXedge) {
      moveCursorL(mouseY);
    }
  }
  centering = true;

  if (mouseY < sizeYedge) {
    constrainCursorUp(mouseX);
  }
  if (mouseY > sizeY-sizeYedge) {
    constrainCursorDown(mouseX);
  }
}

void setCursor(int x, int y) { // move cursor to specified x/y coord
  try {
    Robot robot = new Robot();
    Point p = new Point( x, y );
    SwingUtilities.convertPointToScreen( p, this );
    robot.mouseMove( p.x, p.y );
  }  
  catch (AWTException e) {
  }
}

void  moveCursorL(int y) { // wrap cursor from right to left side of window
  try {
    Robot robotL = new Robot();
    Point pL = new Point( sizeXedge, y );
    SwingUtilities.convertPointToScreen( pL, this );
    robotL.mouseMove( pL.x, pL.y );
  }  
  catch (AWTException e) {
  }
} 

void  moveCursorR(int y) { // wrap cursor from left to right side of window
  try {
    Robot robotR = new Robot();
    Point pR = new Point( sizeX-sizeXedge, y );
    SwingUtilities.convertPointToScreen( pR, this );
    robotR.mouseMove( pR.x, pR.y );
  }  
  catch (AWTException e) {
  }
} 

void  constrainCursorUp(int x) { // constrain y coord of cursor at top of window
  try {
    Robot robotU = new Robot();
    Point pU = new Point( x, sizeYedge );
    SwingUtilities.convertPointToScreen( pU, this );
    robotU.mouseMove( pU.x, pU.y );
  }  
  catch (AWTException e) {
  }
} 

void  constrainCursorDown(int x) { // constrain y coord of cursor at bottom of window
  try {
    Robot robotD = new Robot();
    Point pD = new Point( x, sizeY-sizeYedge );
    SwingUtilities.convertPointToScreen( pD, this );
    robotD.mouseMove( pD.x, pD.y );
  }  
  catch (AWTException e) {
  }
} 
//--------------------------------------//
